/*
 * CrossFire, A Multiplayer game for X-windows
 * 
 * Copyright (C) 2005, 2006, 2007 Marc Lehmann & Crossfire+ Development Team
 * Copyright (C) 2002-2003 Mark Wedel & The Crossfire Development Team
 * Copyright (C) 1992 Frank Tore Johansen
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author can be reached via e-mail to <crossfire@schmorp.de>
 */

/**
 * \file
 * Main client/server loops.
 *
 * \date 2003-12-02
 *
 * loop.c mainly deals with initialization and higher level socket
 * maintanance (checking for lost connections and if data has arrived.)
 */

#define BG_SCRUB_RATE 4 // how often to send a face in the background

#include <global.h>
#include <sproto.h>
#include <sockproto.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include <unistd.h>
#include <arpa/inet.h>

#include <loader.h>

#define MAX_QUEUE_DEPTH   500 //TODO
#define MAX_QUEUE_BACKLOG 3.  //TODO

void
client::reset_state ()
{
  if (!pl)
    return;

  pl->run_on  = 0;
  pl->fire_on = 0;
}

void
client::queue_command (packet_type *handler, char *data, int datalen)
{
  tstamp stamp = now ();

  if (cmd_queue.size () >= MAX_QUEUE_DEPTH)
    {
      //TODO: just disconnect here?
      reset_state ();
      send_packet_printf ("drawinfo %d command queue overflow, ignoring.", NDI_RED);
    }
  else
    {
      cmd_queue.push_back (command ());
      command &cmd = cmd_queue.back ();
      cmd.stamp   = stamp;
      cmd.handler = handler;
      cmd.data    = salloc<char> (datalen + 1, data);
      cmd.datalen = datalen;
    }
}

bool
client::handle_command ()
{
  if (!cmd_queue.empty ()
      && state == ST_PLAYING
      && pl->ob->speed_left > 0.f)
    {
      command &cmd = cmd_queue.front ();

      if (cmd.stamp + MAX_QUEUE_BACKLOG < now ())
        {
          reset_state ();
          send_packet_printf ("drawinfo %d ignoring delayed commands.", NDI_RED);
        }
      else
        execute (cmd.handler, cmd.data, cmd.datalen);

      cmd_queue.pop_front ();
      return true;
    }
  else
    return false;
}

void
flush_sockets (void)
{
  for (sockvec::iterator i = clients.begin (); i != clients.end (); ++i)
    (*i)->flush ();
}

/**
 * This checks the sockets for input, does the right thing.
 *
 * A bit of this code is grabbed out of socket.c
 * There are 2 lists we need to look through - init_sockets is a list
 * 
 */
void
doeric_server (void)
{
#ifdef CS_LOGSTATS
  if ((time (NULL) - cst_lst.time_start) >= CS_LOGTIME)
    write_cs_stats ();
#endif

  //TODO: should not be done here, either
  for (unsigned i = 0; i < clients.size (); ++i)
    {
      client *ns = clients [i];

      ns->tick ();
      ns->refcnt_chk ();
    }
}

void
client::tick ()
{
  if (!pl || destroyed ())
    return;

  /* Update the players stats once per tick.  More efficient than
   * sending them whenever they change, and probably just as useful
   */
  esrv_update_stats (pl);

  if (last_weight != -1 && last_weight != WEIGHT (pl->ob))
    {
      esrv_update_item (UPD_WEIGHT, pl->ob, pl->ob);
      if (last_weight != WEIGHT (pl->ob))
        LOG (llevError, "esrv_update_item(UPD_WEIGHT) did not set player weight: is %lu, should be %lu\n",
             (unsigned long) last_weight, WEIGHT (pl->ob));
    }

  draw_client_map (pl->ob);

  if (update_look)
    esrv_draw_look (pl->ob);

  if (askface.empty () && !partial_face_ofs)
    {
      // regularly send a new face when queue is empty
      if (bg_scrub && !--bg_scrub && enable_bg_scrub)
        while (scrub_idx < faces.size () - 1)
          {
            ++scrub_idx;
            if (!faces_sent [scrub_idx])
              {
                send_face (scrub_idx);
                flush_fx ();

                bg_scrub = 1; // send up to one face per tick, unless an image was requested
                break;
              }
          }

      rate_avail = max_rate - outputbuffer_len ();
    }
  else
    {
      int ol = outputbuffer_len ();

      rate_avail = min (max_rate, rate_avail + max_rate);
      rate_avail -= ol;

      int avail = rate_avail;

      // if we can split images, transfer up to mss packets if possible
      // but never more 768 bytes more.
      if (fxix) avail += min (768, mss - (ol % mss));

      bg_scrub = BG_SCRUB_RATE;

      while (avail > 0)
        {
          if (partial_face_ofs)
            {
              if (facedata *d = face_data (partial_face, faceset))
                {
                  // 9 bytes is enough for fx_FFFOOO, 40 leaves some room
                  int chunk = min (min (avail - 40, MAXSOCKBUF - 9), partial_face_ofs);

                  if (chunk <= 0)
                    break;

                  partial_face_ofs -= chunk;

                  packet sl ("ix");

                  sl << ber32 (partial_face)
                     << ber32 (partial_face_ofs)
                     << data (d->data.data () + partial_face_ofs, chunk);

                  send_packet (sl);
                }
              else
                partial_face_ofs = 0;
            }
          else if (!askface.empty ())
            {
              // use a lifo to send most recently requested images
              faceidx facenum = askface.back ();
              askface.pop_back ();

              if (fxix)
                {
                  if (facedata *d = face_data (facenum, faceset))
                    {
                      partial_face     = facenum;
                      partial_face_ofs = d->data.size ();
                    }
                }
              else
                send_image (facenum);
            }
          else
            break;

          int consumed = ol - outputbuffer_len ();
          avail      -= consumed;
          rate_avail -= consumed;
        }
    }
}

