/*
 * CrossFire, A Multiplayer game
 * 
 * Copyright (C) 2005, 2006, 2007 Marc Lehmann & Crossfire+ Development Team
 * Copyright (C) 1992 Frank Tore Johansen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * The author can be reached via e-mail to mark@pyramid.com
 */
 
/**
 * \file
 * Low-level socket-related functions.
 *
 * \date 2003-12-02
 *
 * Contains some base functions that both the client and server
 * can use.  As such, depending what we are being compiled for will
 * determine what we can include.  the client is designed have
 * CFCLIENT defined as part of its compile flags.
 */

using namespace std;

#include <global.h>
#include <sproto.h>
#include <cstdarg>

#ifdef __linux__
# include <sys/types.h>
# include <sys/socket.h>
# include <netinet/in.h>
# include <netinet/tcp.h>
#endif

// use a really low timeout, as it doesn't cost any bandwidth, and you can
// easily die in 20 seconds...
#define SOCKET_TIMEOUT1 10 * 1000
#define SOCKET_TIMEOUT2 20 * 1000

void
client::flush ()
{
  if (destroyed ())
    return;

#ifdef __linux__
  // check time of last ack, and, if too old, kill connection
  struct tcp_info tcpi;
  socklen_t len = sizeof (tcpi);

  if (!getsockopt (fd, IPPROTO_TCP, TCP_INFO, &tcpi, &len) && len == sizeof (tcpi))
    {
      unsigned int diff = tcpi.tcpi_last_ack_recv - tcpi.tcpi_last_data_sent;

      rtt    = tcpi.tcpi_rtt;
      rttvar = tcpi.tcpi_rttvar;

      if (tcpi.tcpi_unacked && SOCKET_TIMEOUT1 < diff && diff < 0x80000000UL   // ack delayed for 20s
          && SOCKET_TIMEOUT2 < tcpi.tcpi_last_data_sent)       // no data sent for 10s
        {
          LOG (llevDebug, "Connection on fd %d closed due to ack timeout (%u/%u/%u)\n", fd,
               (unsigned) tcpi.tcpi_last_ack_recv, (unsigned) tcpi.tcpi_last_data_sent, (unsigned) tcpi.tcpi_unacked);
          destroy ();
        }
    }
#endif

  /**
   * Writes data to socket.
   *
   * When the socket is clear to write, and we have backlogged data, this
   * is called to write it out.
   */

  if (!outputbuffer.len || socket_ev.poll () & PE_W)
    return;

  write_outputbuffer ();
}

void
client::write_outputbuffer ()
{
  while (outputbuffer.len)
    {
      int res = write (fd, outputbuffer.data + outputbuffer.start,
                       min (outputbuffer.len, SOCKETBUFSIZE - outputbuffer.start));

      if (res > 0)
        {
          outputbuffer.start += res;
          /* wrap back to start of buffer */
          if (outputbuffer.start == SOCKETBUFSIZE)
            outputbuffer.start = 0;

          outputbuffer.len -= res;
#ifdef CS_LOGSTATS
          cst_tot.obytes += res;
          cst_lst.obytes += res;
#endif
        }
      else if (res == 0)
        {
          LOG (llevError, "socket write failed, connection closed.\n");
          destroy ();
          return;
        }
      else if (errno == EINTR)
        {
          // just retry
        }
      else if (errno == EAGAIN)
        {
          // delay till ready
          socket_ev.poll (socket_ev.poll () | PE_W);
          socket_ev.start ();
          return;
        }
      else
        {
          LOG (llevError, "socket write failed: %s\n", strerror (errno));
          destroy ();
          return;
        }
    }

  socket_ev.poll (socket_ev.poll () & ~PE_W);
}

/******************************************************************************
 *
 * Start of read routines.
 *
 ******************************************************************************/

int
client::next_packet ()
{
  if (inbuf_len >= 2)
    {
      int pkt_len = (inbuf [0] << 8) | inbuf [1];

      if (inbuf_len >= 2 + pkt_len)
        return 2 + pkt_len;

      if (inbuf_len == sizeof (inbuf))
        {
          send_packet_printf ("drawinfo %d input buffer overflow - closing connection.", NDI_RED);
          destroy ();
          return -1;
        }
    }

  return 0;
}

void
client::skip_packet (int len)
{
  inbuf_len -= len;
  memmove (inbuf, inbuf + len, inbuf_len);
}

/*****************************************************************************
 * Start of command dispatch area.
 * The commands here are protocol commands. 
 ****************************************************************************/

// SocketCommand, PlayingCommand, should not exist with those ugly casts
#define SC(cb) (void *)static_cast<void (*)(char *, int, client *)>(cb),
#define PC(cb) (void *)static_cast<void (*)(char *, int, player *)>(cb), PF_PLAYER |

/**
 * Dispatch table for the server.
 */
static struct packet_type packets[] = {
  {"ncom",        PC(NewPlayerCmd) PF_PLAYING | PF_COMMAND6 },
  {"command",     PC(PlayerCmd)    PF_PLAYING | PF_COMMAND0 },

  {"examine",     PC(ExamineCmd)   PF_PLAYING },
  {"ex",          PC(ExCmd)        PF_PLAYING },
  {"apply",       PC(ApplyCmd)     PF_PLAYING },
  {"lookat",      PC(LookAt)       PF_PLAYING },
  {"lock",        PC(LockItem)     PF_PLAYING },
  {"mark",        PC(MarkItem)     PF_PLAYING },
  {"move",        PC(MoveCmd)      PF_PLAYING },
  {"ext",         PC(ExtCmd)       0 }, // CF+
  {"mapredraw",   PC(MapRedrawCmd) 0 },
  {"mapinfo",     PC(MapInfoCmd)   0 }, // CF+

  {"reply",       SC(ReplyCmd)     0 },
  {"exti",        SC(ExtiCmd)      0 }, // CF+
  {"addme",       SC(AddMeCmd)     0 },
  {"askface",     SC(AskFaceCmd)   0 },
  {"requestinfo", SC(RequestInfo)  0 },
  {"setfacemode", SC(SetFaceMode)  0 },
  {"setsound",    SC(SetSound)     0 },
  {"setup",       SC(SetUp)        0 },
  {"version",     SC(VersionCmd)   0 },
  {"toggleextendedinfos", SC(ToggleExtendedInfos) 0 }, /*Added: tchize */
  {"toggleextendedtext",  SC(ToggleExtendedText)  0 },   /*Added: tchize */
  {"asksmooth",   SC(AskSmooth)    0 },     /*Added: tchize (smoothing technologies) */
};

bool
client::may_execute (const packet_type *pkt) const
{
  return (!(pkt->flags & PF_PLAYER)  || pl)
      && (!(pkt->flags & PF_PLAYING) || state == ST_PLAYING);
}

// HACK: some commands currently should be executed 
// even when the player is frozen. this hack detects
// those commands. it should be folded into may_execute,
// but kept seperate to emphasise the hack aspect, i.e.
// do it better, then remove.
static bool
always_immediate (const client *ns, const packet_type *pkt, const char *data, int len)
{
  if (!(pkt->flags & (PF_COMMAND0 | PF_COMMAND6)))
    return false;

  if (!ns->pl || !ns->pl->ob || !ns->pl->ob->map)
    return false;

  if (pkt->flags & PF_COMMAND6)
    {
      data += 6;
      len  -= 6;
    }

  if (len > 4 && !strncmp (data, "say " , 4))
    return true;
  if (len > 5 && !strncmp (data, "chat ", 5))
    return true;

  return false;
}

void
client::execute (const packet_type *pkt, char *data, int datalen)
{
  if (may_execute (pkt) || always_immediate (this, pkt, data, datalen))
    {
      //TODO: only one format
      if (pkt->flags & PF_PLAYER)
        ((void (*)(char *, int, player *))pkt->cb)((char *)data, datalen, pl);
      else
        ((void (*)(char *, int, client *))pkt->cb)((char *)data, datalen, this);
    }
  else
    send_packet_printf ("drawinfo %d ERROR: you cannot execute '%s' now.", NDI_RED, pkt->name);
}

bool
client::handle_packet ()
{
  int pkt_len = next_packet ();

  if (!pkt_len)
    return false;
  else if (pkt_len < 0)
    {
      LOG (llevError, "read error on player %s\n",
                      pl && pl->ob ? &pl->ob->name : "[anonymous]");
      destroy ();
      return false;
    }

  inbuf [pkt_len] = 0;      /* Terminate buffer - useful for string data */

  /* First, break out beginning word.  There are at least
   * a few commands that do not have any paremeters.  If
   * we get such a command, don't worry about trying
   * to break it up.
   */
  int datalen;
  char *data = strchr ((char *)inbuf + 2, ' ');

  if (data)
    {
      *data++ = 0;
      datalen = pkt_len - (data - (char *)inbuf);
    }
  else
    {
      data = (char *)inbuf + 2; // better read garbage than segfault
      datalen = 0;
    }

  for (packet_type *pkt = packets; pkt < packets + (sizeof (packets) / sizeof (packets[0])); ++pkt)
    if (!strcmp ((char *)inbuf + 2, pkt->name))
      {
        if (pkt->flags & PF_PLAYER && !always_immediate (this, pkt, data, datalen))
          queue_command (pkt, data, datalen);
        else
          execute (pkt, data, datalen);

        goto next_packet;
      }

  // If we get here, we didn't find a valid command.
  send_packet_printf ("drawinfo %d ERROR: command '%s' not supported.", NDI_RED, (char *)inbuf + 2);
next_packet:
  skip_packet (pkt_len);

  // input buffer has space again
  socket_ev.poll (socket_ev.poll () | PE_R);

  return true;
}

// callback called when socket is either readable or writable
void
client::socket_cb (iow &w, int got)
{
  //TODO remove when we have better socket cleanup logic
  if (destroyed ())
    {
      socket_ev.poll (0);
      return;
    }

  if (got & PE_W)
    {
      write_outputbuffer ();

      if (!outputbuffer.len)
        socket_ev.poll (socket_ev.poll () & ~PE_W);
    }

  if (got & PE_R)
    {
      //TODO: rate-limit tcp connection in better ways, important

      int amount = sizeof (inbuf) - inbuf_len;

      if (!amount)
        {
          // input buffer full
          socket_ev.poll (socket_ev.poll () & ~PE_R);
          return;
        }

      amount = read (fd, inbuf + inbuf_len, amount);

      if (!amount)
        {
          destroy ();
          return;
        }
      else if (amount < 0)
        {
          if (errno != EAGAIN && errno != EINTR)
            {
              LOG (llevError, "read error: %s\n", strerror (errno));
              destroy ();
              return;
            }

          // should not be here, normally
        }
      else
        {
          inbuf_len += amount;

          cst_tot.ibytes += amount;
          cst_lst.ibytes += amount;

          cmd_ev.start ();
        }
    }
}

// called whenever we have additional commands to process
void
client::cmd_cb (iw &w)
{
  if (handle_packet ())
    w.start ();
  else
    flush ();
}

/*******************************************************************************
 *
 * Start of write related routines.
 *
 ******************************************************************************/

/**
 * Adds data to a socket buffer for whatever reason.
 *
 * ns is the socket we are adding the data to, buf is the start of the
 * data, and len is the number of bytes to add.
 */
void
client::send (void *buf_, int len)
{
  char *buf = (char *)buf_;

  if (destroyed () || !buf)
    return;

  if (len + outputbuffer.len > SOCKETBUFSIZE)
    {
      LOG (llevDebug, "socket on fd %d has overrun internal buffer - marking as dead\n", fd);
      // shutdown the socket, this is safer than destroying it immediately
      // as lots of code in the callchain might still access the map etc.
      shutdown (fd, SHUT_RDWR);
      return;
    }

  int avail, end;

  /* data + end is where we start putting the new data.  The last byte
   * currently in use is actually data + end -1
   */
  end = outputbuffer.start + outputbuffer.len;
  /* The buffer is already in a wrapped state, so adjust end */
  if (end >= SOCKETBUFSIZE)
    end -= SOCKETBUFSIZE;

  avail = SOCKETBUFSIZE - end;

  /* We can all fit it behind the current data without wrapping */
  if (avail >= len)
    memcpy (outputbuffer.data + end, buf, len);
  else
    {
      memcpy (outputbuffer.data + end, buf, avail);
      memcpy (outputbuffer.data, buf + avail, len - avail);
    }

  outputbuffer.len += len;
}

/**
 * Takes a string of data, and writes it out to the socket. A very handy
 * shortcut function.
 */
void
client::send_packet (packet &sl)
{
  if (destroyed ())
    return;

  if (sl.length () >= MAXSOCKBUF)
    {
      LOG (llevError, "Trying to send a buffer beyond properly size, len =%d\n", sl.length ());
      /* Almost certainly we've overflowed a buffer, so quit now to make
       * it easier to debug.
       */
      abort ();
    }

  if (!sl.length ())
    return;

  assert (sl.hdrlen == 2);

  sl.buf_ [0] = sl.length () >> 8;
  sl.buf_ [1] = sl.length ()     ;

  send (sl.buf_, sl.length () + sl.hdrlen);
}

void
client::send_packet (const char *buf, int len)
{
  packet sl;

  sl << data (buf, len);
  send_packet (sl);
}

void
client::send_packet (const char *buf)
{
  send_packet (buf, strlen (buf));
}

void
client::send_packet_printf (const char *format, ...)
{
  packet sl;

  va_list ap;
  va_start (ap, format);
  sl.vprintf (format, ap);
  va_end (ap);

  send_packet (sl);
}

void
client::send_drawinfo (const char *msg, int flags)
{
  send_packet_printf ("drawinfo %d %s", flags, msg);
}

void
client::send_msg (int color, const char *type, const char *msg)
{
  if (can_msg)
    send_packet_printf ("msg %d %s %s", color, type, msg);
  else if (color < 0)
    return; // client cannot handle this
  else if (strchr (msg, '<') || strchr (msg, '&'))
    {
      //TODO: should escape/modify to old syntax
      send_packet_printf ("drawinfo %d %s", color, msg);
    }
  else
    send_packet_printf ("drawinfo %d %s", color, msg);
}

/***********************************************************************
 *
 * packet functions/utilities
 *
 **********************************************************************/

packet::packet (const char *name)
{
  reset ();

  int len = strlen (name);
  memcpy (cur, name, len); cur += len;
  *cur++ = ' ';
}

packet &packet::operator <<(const ber32 v)
{
  enum { maxlen = 32 / 7 + 1};
  uint8 buf[maxlen];
  uint8 *p = buf + maxlen;
  uint32 val = v.val;

  *--p = val & 0x7F;

  while (val > 0x7F)
    {
      val >>= 7;
      *--p = (val & 0x7F) | 0x80;
    }

  return *this << data (p, buf + maxlen - p);
}

packet &packet::operator <<(const data   &v)
{
  if (room () < v.len)
    reset ();
  else
    {
      if (v.len)
        {
          memcpy (cur, v.ptr, v.len);
          cur += v.len;
        }
    }

  return *this;
}

packet &packet::operator <<(const data8  &v)
{
  unsigned int len = min (v.len, 0x00FF);
  return *this << uint8  (len) << data (v.ptr, len);
}

packet &packet::operator <<(const data16 &v)
{
  unsigned int len = min (v.len, 0xFFFF);
  return *this << uint16 (len) << data (v.ptr, len);
}

packet &packet::operator <<(const char *v)
{
  return *this << data (v, strlen (v ? v : 0));
}

void
packet::vprintf (const char *format, va_list ap)
{
  int size = room ();

  int len = vsnprintf ((char *)cur, size, format, ap);

  if (len >= size)
    return reset ();

  cur += len;
}

/******************************************************************************
 *
 * statistics logging functions.
 *
 ******************************************************************************/

#ifdef CS_LOGSTATS

/* cst_tot is for the life of the server, cst_last is for the last series of
 * stats
 */
CS_Stats cst_tot, cst_lst;

/**
 * Writes out the gathered stats.  We clear cst_lst.
 */
void
write_cs_stats (void)
{
  time_t now = time (NULL);

  /* If no connections recently, don't both to log anything */
  if (cst_lst.ibytes == 0 && cst_lst.obytes == 0)
    return;

  /* CSSTAT is put in so scripts can easily find the line */
  LOG (llevInfo, "CSSTAT: %.16s tot %d %d %d %d inc %d %d %d %d\n",
       ctime (&now), cst_tot.ibytes, cst_tot.obytes, cst_tot.max_conn,
       now - cst_tot.time_start, cst_lst.ibytes, cst_lst.obytes, cst_lst.max_conn, now - cst_lst.time_start);
  cst_lst.ibytes = 0;
  cst_lst.obytes = 0;
  cst_lst.time_start = now;
}
#endif

